# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Query environment variable for documentation building."""
import os


def _import_tfg_docs():
  """Checks if __init__.py imports should be executed (for buildling docs)."""
  return os.getenv("TFG_DOC_IMPORTS", "0") == "1"


def enable_tfg_doc_imports():
  """Re-enables the imports in the __init__.py so that docs can be built."""
  os.environ["TFG_DOC_IMPORTS"] = "1"
